<?php

declare(strict_types=1);

namespace Drupal\file_visibility;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\StreamWrapper\StreamWrapperManager;
use Drupal\file\FileInterface;
use Drupal\file\FileRepositoryInterface;

/**
 * Default implementation for the module's main service.
 */
class FileVisibility implements FileVisibilityInterface {

  /**
   * The entities to be processed.
   *
   * @var list<\Drupal\Core\Entity\FieldableEntityInterface>
   */
  protected array $entities = [];

  public function __construct(
    protected readonly FileVisibilityPluginManager $manager,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly FileSystemInterface $fileSystem,
    protected readonly FileRepositoryInterface $fileRepository,
    protected readonly AccountInterface $currentUser,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function updateFilesVisibility(EntityInterface $entity): void {
    if ($entity instanceof FieldableEntityInterface && !$entity instanceof FileInterface) {
      // Queue the entity but also the files used by this entity to be processed
      // on the kernel terminate phase for performance reasons. The entity is
      // limited to types that can determine the file usages. Source entities
      // are included (e.g., node), but also traversable entities, such as
      // media or paragraph.
      $this->entities[] = [$entity, $this->getEntityUsedFiles($entity)];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function destruct(): void {
    while ($this->entities) {
      [$entity, $initialFileIds] = array_shift($this->entities);
      $this->processEntity($entity, $initialFileIds);
    }
  }

  /**
   * Sets the entity files visibility.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity.
   * @param array<int|string, int|string> $initialFileIds
   *   IDs of the files used by the source entity.
   */
  protected function processEntity(FieldableEntityInterface $entity, array $initialFileIds): void {
    // Make sure the access control handler static cache is not stale.
    $this->entityTypeManager->getAccessControlHandler($entity->getEntityTypeId())->resetCache();

    // Other processes might have changed the files used by this source entity,
    // so we need to re-check the files. Merge the initial files with the
    // current ones because some files might have been added and others removed.
    $fids = array_unique([...$initialFileIds, ...$this->getEntityUsedFiles($entity)]);
    if (!$fids) {
      return;
    }

    $anonymous = new AnonymousUserSession();
    $fileStorage = $this->entityTypeManager->getStorage('file');
    foreach ($fileStorage->loadMultiple($fids) as $file) {
      $hasPublicAccess = !$this->getFileAccess($file, $anonymous)->isForbidden();

      // @todo What about `$settings['file_additional_public_schemes']`? With
      //   the current architecture this is problematic because we'll need to
      //   know/store the public scheme, when switching back from private.
      if (!$hasPublicAccess && str_starts_with($file->getFileUri(), 'public://')) {
        $this->moveFile($file, static::PRIVATE_LOCATION . StreamWrapperManager::getTarget($file->getFileUri()));
      }
      elseif ($hasPublicAccess && str_starts_with($file->getFileUri(), static::PRIVATE_LOCATION)) {
        [, $path] = explode('/', StreamWrapperManager::getTarget($file->getFileUri()), 2);
        $this->moveFile($file, "public://$path");
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityUsedFiles(FieldableEntityInterface $entity): array {
    $fids = [];

    // Collect files from plugins.
    foreach ($this->manager->getPlugins() as $plugin) {
      foreach ($plugin->getEntityFiles($entity) as $fid) {
        $fids[$fid] = $fid;
      }
    }

    return array_values($fids);
  }

  /**
   * {@inheritdoc}
   */
  public function getFileAccess(FileInterface $file, AccountInterface $account): AccessResultInterface {
    $access = FALSE;
    $metadata = (new CacheableMetadata())->addCacheableDependency($file);
    $processed = [];

    $paths = [];
    foreach ($this->manager->getPlugins() as $plugin) {
      foreach ($plugin->getPathsToFile($file) as $path) {
        $paths[] = $path;
      }
    }

    if (!$paths) {
      // This is a file not referred to by any entity. It might be a temporary
      // file after an upload on an entity addition form when the entity hasn't
      // been yet saved. In this case only the file's author is permitted, and
      // we cannot make an opinion. Or it might be a permanent file with all its
      // referrers were deleted.
      return AccessResult::forbiddenIf($file->isPermanent() || $file->getOwnerId() != $this->currentUser->id())
        ->addCacheableDependency($metadata);
    }

    foreach ($paths as $path) {
      $pathAccess = TRUE;
      foreach ($path as $entity) {
        assert($entity instanceof FieldableEntityInterface);
        // Avoid useless calls to $entity->access().
        $revisionId = $entity instanceof RevisionableInterface ? $entity->getRevisionId() : '0';
        if ($pathAccess && !isset($processed[$entity->getEntityTypeId()][$entity->id()][$revisionId])) {
          $pathAccess = $entity->access('view', $account);
          $metadata->addCacheableDependency($entity);
          $processed[$entity->getEntityTypeId()][$revisionId] = $pathAccess;
        }
      }

      // Once $access is TRUE, we don't change it, but we still iterate over all
      // remaining entities to collect cacheable metadata.
      $access = $access ?: $pathAccess;
    }

    return AccessResult::forbiddenIf(!$access)->addCacheableDependency($metadata);
  }

  /**
   * Moves a given file to a given destination.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file entity.
   * @param string $destinationUri
   *   The destination URI.
   */
  protected function moveFile(FileInterface $file, string $destinationUri): void {
    $destinationDir = $this->fileSystem->dirname($destinationUri);
    if (!$this->fileSystem->prepareDirectory($destinationDir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      throw new \RuntimeException("Directory $destinationDir cannot be created or it's not writable.");
    }
    $this->fileRepository->move($file, $destinationUri, FileExists::Replace);
  }

}
